from py2neo import Subgraph
import networkx
from NeoMetaTracker.visualizer._base_visualizer import BaseVisualizer
import matplotlib.pyplot as plt


# WIP: not usable atm


class NetworkxVisualizer:
    def create_schema_png(self):
        nx_graph = self._py2neo_subgraph_to_networkxgraph(
            py2neo_subgraph=self.neo_meta_logger.capture_points[0].schema
        )

    def _py2neo_subgraph_to_networkxgraph(
        self,
        py2neo_subgraph: Subgraph,
    ) -> networkx.MultiDiGraph:
        nx_graph: networkx.MultiDiGraph = networkx.MultiDiGraph()
        for node in py2neo_subgraph.nodes:
            nx_graph.add_node(
                node["__label_name"],
                labels=list[node.labels],
                properties=dict(node),
            )
        for rel in py2neo_subgraph.relationships:
            nx_graph.add_edge(
                rel.start_node["__label_name"],
                rel.end_node["__label_name"],
                key=rel.type,
                type=rel.type,
                properties=dict(rel),
            )
        return nx_graph



from __future__ import annotations
from typing import Union, List, Dict, Generic, TYPE_CHECKING


if TYPE_CHECKING:
    from NeoMetaTracker.capture_point import CapturePoint
from pathlib import Path
import py2neo

from NeoMetaTracker.visualizer._base_visualizer import BaseVisualizer


class GraphSchema(py2neo.Subgraph):
    parent_capture_point: CapturePoint

    def visualize(
        self,
        visualizer_class: BaseVisualizer,
        to_file: Union[Path, str] = None,
    ):
        visualizer: BaseVisualizer = visualizer_class(self)
        if to_file:
            visualizer.generate_file(to_file)
        else:
            return visualizer.generate_object()

    @classmethod
    def from_neo4j_schema_vis_data(
        cls,
        neo4j_schema_vis_data: Dict[str, List[Union[py2neo.Node, py2neo.Relationship]]],
        parent_capture_point: CapturePoint,
        extra_props: Dict = None,
    ) -> "GraphSchema":
        # ToDo: Refactor or break this function down. Too complex/spaghetti!

        # the nodes are bound atm. meaning they belong to a certain DB. we need to unbound them.
        # we clean the nodes and relationship from Neo4j's `call db.schema.visualization` from IDs (the IDs are random anyway on every call and consequently worthless outside of the own transaction)
        # We connect the nodes and relationships into a py2neo.Subgraph
        # store that in an instance of GraphSchema
        schema_graph: GraphSchema = cls()
        if not parent_capture_point.parent_logger.all_schema_nodes:
            parent_capture_point.parent_logger.all_schema_nodes = {}
        if not parent_capture_point.parent_logger.all_schema_rels:
            parent_capture_point.parent_logger.all_schema_rels = {}
        if not extra_props:
            extra_props = {}
        extra_props = extra_props | {"__neo_meta_logger_node": True}
        # copy all nodes to track if they are involed in a relation
        nodes_without_relation = list(neo4j_schema_vis_data["nodes"])
        for rel in neo4j_schema_vis_data["relationships"]:
            if parent_capture_point.parent_logger.rel_filter_func(rel) is None:
                continue
            rel_nodes = [None, None]
            for node in neo4j_schema_vis_data["nodes"]:
                # we allready fiund start and end node for our relation. no need to search further
                if rel_nodes[0] and rel_nodes[1]:
                    break
                # Check if schema-node is "unwanted"/filtered-out
                if parent_capture_point.parent_logger.node_filter_func(node) is None:
                    if node in nodes_without_relation:
                        nodes_without_relation.remove(node)
                    continue
                # check if schema-node is a "zero change" (e.g. just an index on the label was created) and if zero changes are unwanted
                if (
                    list(node.labels)[0] not in parent_capture_point.labels_count
                    and parent_capture_point.parent_logger.ignore_zero_changes
                ):
                    if node in nodes_without_relation:
                        nodes_without_relation.remove(node)
                    continue

                if node.identity in [rel.start_node.identity, rel.end_node.identity]:
                    if node in nodes_without_relation:
                        nodes_without_relation.remove(node)
                    clean_rel_node = cls._get_or_create_unbound_schema_node(
                        node=node,
                        parent_capture_point=parent_capture_point,
                        extra_props=extra_props,
                    )
                    if rel.start_node.identity == rel.end_node.identity:
                        rel_nodes[0] = rel_nodes[1] = clean_rel_node
                    elif node.identity == rel.start_node.identity:
                        rel_nodes[0] = clean_rel_node
                    elif node.identity == rel.end_node.identity:
                        rel_nodes[1] = clean_rel_node
            if rel_nodes[0] and rel_nodes[1]:
                rel_ident = f"{list(rel_nodes[0].labels)[0]}_{type(rel).__name__}_{list(rel_nodes[1].labels)[0]}"
                if rel_ident in parent_capture_point.parent_logger.all_schema_rels:
                    clean_rel = parent_capture_point.parent_logger.all_schema_rels[
                        rel_ident
                    ]
                else:
                    parent_capture_point.parent_logger.all_schema_rels[
                        rel_ident
                    ] = clean_rel = py2neo.Relationship(
                        rel_nodes[0],
                        type(rel).__name__,
                        rel_nodes[1],
                        **extra_props,
                    )
                schema_graph = schema_graph | clean_rel

        # atm we ignored any nodes without any relation.
        # lets make up that leeway
        single_nodes_cleaned = []

        for node in nodes_without_relation:

            if not parent_capture_point.parent_logger.node_filter_func(node):
                continue
            if (
                list(node.labels)[0] not in parent_capture_point.labels_count
                and parent_capture_point.parent_logger.ignore_zero_changes
            ):
                continue
            clean_node = cls._get_or_create_unbound_schema_node(
                node=node,
                parent_capture_point=parent_capture_point,
                extra_props=extra_props,
            )
            single_nodes_cleaned.append(
                cls._get_or_create_unbound_schema_node(
                    node=node,
                    parent_capture_point=parent_capture_point,
                    extra_props=extra_props,
                )
            )
        # hardcoded class name in subgraph. we need to workaround this. can be removed with https://github.com/py2neo-org/py2neo/issues/940 merged
        # original line: return schema_graph | cls(single_nodes_cleaned)

        # workaround:
        sb: py2neo.Subgraph = schema_graph | cls(single_nodes_cleaned)
        gs = GraphSchema(nodes=sb.nodes, relationships=sb.relationships)
        gs.parent_capture_point = parent_capture_point
        return gs

    @classmethod
    def _get_or_create_unbound_schema_node(
        cls,
        node: py2neo.Node,
        parent_capture_point: "CapturePoint",
        extra_props: Dict = None,
    ) -> py2neo.Node:
        if tuple(
            node.labels
        ) not in parent_capture_point.parent_logger.all_schema_nodes or (
            parent_capture_point.parent_logger.all_schema_nodes[tuple(node.labels)][
                "__node_count"
            ]
            != parent_capture_point.labels_count[tuple(node.labels)[0]]
        ):
            clean_node = py2neo.Node(
                *list(node.labels),
                **(
                    extra_props
                    | {
                        "__label_name": ":".join(list(node.labels)),
                        "__node_count": parent_capture_point.labels_count[
                            tuple(node.labels)[0]
                        ],
                    }
                ),
            )
            clean_node.__primarykey__ = "__label_name"
            clean_node.__primarylabel__ = list(node.labels)[0]

            parent_capture_point.parent_logger.all_schema_nodes[
                tuple(node.labels)
            ] = clean_node

        else:
            clean_node = parent_capture_point.parent_logger.all_schema_nodes[
                tuple(node.labels)
            ]
        return clean_node



from __future__ import annotations
from typing import Dict, TYPE_CHECKING
from NeoMetaTracker.graph_scheme import GraphSchema
import time

if TYPE_CHECKING:
    from NeoMetaTracker.neo_meta_logger import NeoMetaTracker


class CapturePoint:
    """The state of a Neo4j database in terms of relationship and node count at a certain time"""

    def __init__(
        self,
        parent_logger: NeoMetaTracker,
        name: str,
        labels_count: Dict[str, int],
        relations_count: Dict[str, int],
        neo4j_schema_vis_data: Dict,
    ):
        self.parent_logger = parent_logger
        self.name = name
        self.timestamp: float = time.time()
        self.labels_count = labels_count
        self.relations_count = relations_count
        self.schema: GraphSchema = GraphSchema.from_neo4j_schema_vis_data(
            neo4j_schema_vis_data,
            parent_capture_point=self,
            extra_props={"capture_name": name} if name else {},
        )

    def __hash__(self):
        return hash(self.timestamp)

    def __eq__(self, other: "CapturePoint") -> bool:
        return type(other) is type(self) and self.timestamp == other.timestamp



from __future__ import annotations
from typing import List, TYPE_CHECKING

import py2neo
from NeoMetaTracker.visualizer._base_visualizer import BaseVisualizer

if TYPE_CHECKING:
    from NeoMetaTracker.graph_scheme import GraphSchema

# sudo apt-get install graphviz graphviz-dev
# pip install pygraphviz
import pygraphviz as pgv


class GraphvizVisualizer(BaseVisualizer):
    def generate_file(self, path):
        self._py2neo_subgraph_to_graphviz(
            neometatracker_graphschemas=self.neo_meta_logger_graph_schemas
        ).write(path)

    def generate_object(self) -> bytes:
        return (
            self._py2neo_subgraph_to_graphviz(
                neometatracker_graphschemas=self.neo_meta_logger_graph_schemas
            )
            .string()
            .encode("utf-8")
        )

    def _py2neo_subgraph_to_graphviz(
        self,
        neometatracker_graphschemas: List[GraphSchema],
    ) -> pgv.AGraph:
        # https://pygraphviz.github.io/documentation/stable/tutorial.html#graphs
        graph: pgv.AGraph = pgv.AGraph(strict=False, directed=True, landscape=False)
        node_track_list: List[py2neo.Node] = []
        for nmt_graphschema in neometatracker_graphschemas:
            if len(neometatracker_graphschemas) == 1:
                graphviz_subg = graph
            else:
                sb_name = (
                    nmt_graphschema.parent_capture_point.name
                    if nmt_graphschema.parent_capture_point.name
                    else ""
                )
                graphviz_subg = graph.add_subgraph(
                    name=sb_name,
                    label=sb_name,
                    cluster=True,
                    style="filled",
                    color="#E0E0E0",
                )
            for node in nmt_graphschema.nodes:
                if node not in node_track_list:
                    node_track_list.append(node)
                    graphviz_subg.add_node(node["__label_name"], **dict(node))

            for rel in nmt_graphschema.relationships:
                graph.add_edge(
                    rel.start_node["__label_name"],
                    rel.end_node["__label_name"],
                    fontsize=8,
                    minlen=2,
                    # key=type(rel).__name__,
                    label=type(rel).__name__ + "   ",
                )
        return graph



import sys
import os
from typing import Dict, List
import json
import py2neo
from DZDutils.neo4j import wait_for_db_boot

if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )
    SCRIPT_DIR = os.path.join(SCRIPT_DIR, "..")
    sys.path.insert(0, os.path.normpath(SCRIPT_DIR))

from NeoMetaTracker import NeoMetaTracker
from NeoMetaTracker.visualizer import GraphvizVisualizer

NEO4J: Dict = json.loads(os.getenv("NEO4J", "{}"))
wait_for_db_boot(NEO4J)

current_dir = os.path.dirname(__file__)
graph = py2neo.Graph(**NEO4J)

graph.run("CREATE OR REPLACE DATABASE schematest")
graph.run("CREATE OR REPLACE DATABASE test")
test_graph = py2neo.Graph(**(NEO4J | {"name": "test"}))
# test_graph = py2neo.Graph(**(NEO4J))
mlog = NeoMetaTracker(test_graph)
mlog.ignore_zero_changes = True

mlog.ignore_labels = ["_SystemNode"]

mlog.capture(name="Init")
test_graph.run(
    'CALL db.index.fulltext.createNodeIndex("Alien_index",["Alien"],["name"])'
)
mlog.capture("indexCreation")
test_graph.run("CREATE (:Alien {name: 'E.T.'})")
mlog.capture("AlienCluster")
test_graph.run("CREATE (:Human{name:'Amina Okujewa'})")
test_graph.run("CREATE (as:Human{name:'Aaron Swartz'})")
mlog.capture(name="HumansCluster")
test_graph.run("CREATE (:Cat{name:'Grumpy Cat'})")
test_graph.run("CREATE (:Pig{name:'Miss Piggy'})")
mlog.capture("AnimalCluster")
test_graph.run("CREATE (:World {name: 'Earth'})")
test_graph.run("CREATE (:World {name: 'Internet'})")
test_graph.run("CREATE (:World{name:'House'})")
mlog.capture("WorldCluster")
test_graph.run(
    "MATCH (wE:World{name:'Earth'}),(ao:Human{name:'Amina Okujewa'}) CREATE (ao)-[:LIVES_ON]->(wE)"
)
test_graph.run(
    "MATCH (wE:World{name:'Earth'}),(c:Cat{name:'Grumpy Cat'}) CREATE (c)-[:LIVES_ON]->(wE)"
)
test_graph.run(
    "MATCH (wI:World{name:'Internet'}),(wE:World{name:'Earth'}),(as:Human{name:'Aaron Swartz'}) CREATE (as)-[:LIVES_ON]->(wI), (as)-[:LIVES_ON]->(wE)"
)
test_graph.run(
    "MATCH (wI:World{name:'Internet'}),(wE:World{name:'Earth'}) CREATE (wI)-[:EXISTS_ON]->(wE)"
)
test_graph.run(
    "MATCH (wH:World{name:'House'}),(wE:World{name:'Earth'}) CREATE (wH)-[:EXISTS_ON]->(wE)"
)
test_graph.run(
    "MATCH (wH:World{name:'House'}),(wE:World{name:'Earth'}),(mp:Pig{name:'Miss Piggy'}) CREATE (mp)-[:LIVES_ON]->(wH),(mp)-[:LIVES_ON]->(wE)"
)
mlog.capture(name="RelationCluster")
test_graph.run("CREATE (:_SystemNode{name:'test'})")
test_graph.run("CREATE (:_SystemNode{name:'test2'})")
test_graph.run(
    "MATCH (sn:_SystemNode{name:'test'}),(sn2:_SystemNode{name:'test2'}) CREATE (sn)-[:WORKS_WITH]->(sn2)"
)
test_graph.run(
    "MATCH (sn:_SystemNode{name:'test'}),(wE:World{name:'Earth'}) CREATE (sn)-[:WATCHES]->(wH)"
)
mlog.capture()
print("###get_numeric_last_changes###\n", mlog.get_numeric_last_changes())
print("###get_schemagraph_last_changes###\n", mlog.get_schemagraph_last_changes())
print(
    "###schema.visualize(GraphvizVisualizer)###\n",
    mlog.capture_points[-1].schema.visualize(GraphvizVisualizer),
)
print(
    mlog.get_schemagraph_last_changes().visualize(
        GraphvizVisualizer, to_file=os.path.join(current_dir, "last_change.dot")
    )
)
print(mlog.visualize(GraphvizVisualizer, to_file=os.path.join(current_dir, "all.dot")))
schema_graph = py2neo.Graph(**(NEO4J | {"name": "schematest"}))
schema_graph.merge(mlog.get_schemagraph_last_changes())



from __future__ import annotations
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from NeoMetaTracker.graph_scheme import GraphSchema


class BaseVisualizer:
    def __init__(self, neo_meta_logger_graph_schemas: List[GraphSchema]):
        if not isinstance(neo_meta_logger_graph_schemas, list):
            self.neo_meta_logger_graph_schemas = [neo_meta_logger_graph_schemas]
        else:
            self.neo_meta_logger_graph_schemas = neo_meta_logger_graph_schemas

    def generate_file(self, path):
        raise NotImplementedError

    def generate_object(self) -> bytes:
        raise NotImplementedError



from typing import Union, Dict, List, Generic
import time
import py2neo
from pathlib import Path
from NeoMetaTracker.capture_point import CapturePoint
from NeoMetaTracker.graph_scheme import GraphSchema
from NeoMetaTracker.visualizer._base_visualizer import BaseVisualizer


class NeoMetaTracker:
    def __init__(self, connection: Union[py2neo.Graph, Dict]):
        self.capture_points: List[CapturePoint] = []
        self.all_schema_nodes: Dict[tuple, py2neo.Node] = None
        self.all_schema_rels: Dict[str, py2neo.Relationship] = None
        self.ignore_labels: List[str] = []
        self.ignore_reliations_types: List[str] = []
        # by zero changes we mean, changes where the node count did not change. e.g. if we create an index for label "myLabel" we also detect a schema change for Label "myLabel"
        # this can be undesirable in some situations. to ignore such changes set `ignore_zero_changes` to True
        self.ignore_zero_changes: bool = False
        if isinstance(connection, dict):
            self.graph: py2neo.Graph = py2neo.Graph(**connection)
        elif isinstance(connection, py2neo.Graph):
            self.graph = connection
        else:
            raise TypeError(
                f"Expected 'py2neo.Graph' or 'dict'. Got '{type(connection)}'"
            )

    def capture(self, name: str = None):
        labels_count = self._get_labels_count()
        relations_count = self._get_relations_count()

        self.capture_points.append(
            CapturePoint(
                name=name,
                parent_logger=self,
                labels_count=labels_count,
                relations_count=relations_count,
                neo4j_schema_vis_data=self._get_neo4j_schema(),
            )
        )

    def node_filter_func(self, node: py2neo.Node):
        """This function can be overiden to filter specifics nodes out of the schema tracker

        example; Remove all schema-nodes with label starting with "_":
            mlog = NeoMetaTracker(test_graph)

            def node_filter(node: py2neo.Node):
                node_label = list(node.labels)[0]
                if node_label.startswith("_") or node_label in self.ignore_labels:
                    return None
                return node

            mlog.node_filter_func = node_filter

        """
        if list(node.labels)[0] in self.ignore_labels:
            return None
        return node

    def rel_filter_func(self, rel: py2neo.Relationship):
        """This function can be overiden to filter specifics nodes out of the schema tracker

        example; Remove all schema-nodes with label starting with "_":
            mlog = NeoMetaTracker(test_graph)

            def my_filter_func(self, rel: py2neo.Relationship):
                if type(rel).__name__.startwith("_") or type(rel).__name__ in self.ignore_reliations_types:
                    return None
            return rel

            mlog.rel_filter_func = my_filter_func

        """
        if type(rel).__name__ in self.ignore_reliations_types:
            return None
        return rel

    def visualize(
        self, visualizer_class: BaseVisualizer, to_file: Union[str, Path] = None
    ):
        schemas_changes = []
        for index, to_cp in enumerate(self.capture_points):
            if index < 1:
                continue
            from_cp = self.capture_points[index - 1]
            changes = self.get_schemagraph_changes(
                from_capture=from_cp, to_capture=to_cp
            )
            changes.parent_capture_point = to_cp
            schemas_changes.append(changes)
        visualizer: BaseVisualizer = visualizer_class(schemas_changes)
        if to_file:
            visualizer.generate_file(to_file)
        else:
            return visualizer.generate_object()

    def _get_neo4j_schema(self):
        # call db.schema.visualization
        return self.graph.run(
            "call db.schema.visualization yield nodes, relationships"
        ).data()[0]

    def _get_labels_count(self) -> Dict[str, int]:
        all_labels: List[str] = self.graph.run(
            "CALL db.labels() yield label return collect(label) as res"
        ).data()[0]["res"]
        labels_count: Dict[str, int] = {}
        for label in all_labels:
            if label in self.ignore_labels:
                continue
            query_label_count = f"""
                    MATCH (n:{label})
                    RETURN count(n) AS res
                    """
            res = self.graph.run(query_label_count).data()[0]["res"]
            if (self.ignore_zero_changes and res != 0) or not self.ignore_zero_changes:
                labels_count[label] = res
        return labels_count

    def _get_relations_count(self) -> Dict[str, int]:
        all_rels: List[str] = self.graph.run(
            "CALL db.relationshipTypes() yield relationshipType return collect(relationshipType) as res"
        ).data()[0]["res"]
        rels_count: Dict[str, int] = {}
        for rel in all_rels:
            if rel in self.ignore_reliations_types:
                continue
            query_rel_count = f"""
                    MATCH ()-[:{rel}]->() 
                    return count(*) AS res
                    """
            res = self.graph.run(query_rel_count).data()[0]["res"]
            if (self.ignore_zero_changes and res != 0) or not self.ignore_zero_changes:
                rels_count[rel] = res
        return rels_count

    def get_capture_point_by(
        self,
        capture_point_index: int = None,
        capture_point_time: float = None,
        capture_point_name: str = None,
    ):
        if capture_point_time:
            return next(
                cp for cp in self.capture_points if capture_point_time == cp.timestamp
            )
        elif capture_point_index:
            return self.capture_points[capture_point_index]
        elif capture_point_name:
            return next(
                cp for cp in self.capture_points if capture_point_name == cp.name
            )

    def get_schemagraph_last_changes(self) -> GraphSchema:
        return self.get_schemagraph_changes_since(
            self.get_capture_point_by(capture_point_index=-2)
        )

    def get_schemagraph_changes_since(
        self,
        capture_point: CapturePoint = None,
    ) -> GraphSchema:
        current_capture_point: CapturePoint = (
            self.capture_points[-1] if self.capture_points else None
        )
        return self.get_schemagraph_changes(
            from_capture=capture_point, to_capture=current_capture_point
        )

    def get_schemagraph_changes(
        self, from_capture: CapturePoint, to_capture: CapturePoint
    ) -> GraphSchema:
        # hardcoded class name in subgraph. we need to workaround this. can be removed with https://github.com/py2neo-org/py2neo/issues/940 merged
        # original line: return from_capture.schema - to_capture.schema
        # workaround:
        sb: py2neo.Subgraph = to_capture.schema - from_capture.schema
        return GraphSchema(nodes=sb.nodes, relationships=sb.relationships)

    def get_numeric_last_changes(self) -> Dict[str, Dict[str, int]]:
        """Get changes, in terms of quantity, of labels and relations since the last capture compared to the current capture

        Returns:
            Dict[str,Dict[str,int]]: A base dictonary "{'labels':{...},'relation':{...}}" containing two dictoniries listing changes (in terms of quantity) for labels and relations.
        """
        return self.get_numeric_changes_since(
            self.get_capture_point_by(capture_point_index=-2)
        )

    def get_numeric_changes_since(
        self,
        capture_point: CapturePoint = None,
    ) -> Dict[str, Dict[str, int]]:
        """Get changes, in terms of quantity, of labels and relations since the last capture compared to the current capture

        Returns:
            Dict[str,Dict[str,int]]: A base dictonary "{'labels':{...},'relation':{...}}" containing two dictoniries listing changes (in terms of quantity) for labels and relations.
        """
        current_capture_point: CapturePoint = (
            self.capture_points[-1] if self.capture_points else None
        )
        return self.get_numeric_changes(
            from_capture=capture_point, to_capture=current_capture_point
        )

    def get_numeric_changes(
        self,
        from_capture: CapturePoint,
        to_capture: CapturePoint,
        ignore_zeros: bool = None,
    ):
        if ignore_zeros is None:
            ignore_zeros = self.ignore_zero_changes
        if len(self.capture_points) <= 1:
            raise ValueError(
                f"You need at least 2 capture points to compare any changes. Got only {len(self.capture_points)} points"
            )
        # Compare labels
        result: Dict = {"labels": {}, "relations": {}}
        if to_capture.labels_count or from_capture.labels_count:
            diff_labels = set(to_capture.labels_count.items()) ^ set(
                from_capture.labels_count.items()
            )
            for diff_label in set([lbl[0] for lbl in diff_labels]):
                res = (
                    to_capture.labels_count[diff_label]
                    if diff_label in to_capture.labels_count
                    else 0
                ) - (
                    from_capture.labels_count[diff_label]
                    if diff_label in from_capture.labels_count
                    else 0
                )
                if (res != 0 and ignore_zeros) or not ignore_zeros:
                    result["labels"][diff_label] = res
        # Compare relations
        if not to_capture.relations_count and not from_capture.relations_count:
            return result
        diff_relations = set(to_capture.relations_count.items()) ^ set(
            from_capture.relations_count.items()
        )
        for diff_rel in set([lbl[0] for lbl in diff_relations]):
            res = (
                to_capture.relations_count[diff_rel]
                if diff_rel in to_capture.relations_count
                else 0
            ) - (
                from_capture.relations_count[diff_rel]
                if diff_rel in from_capture.relations_count
                else 0
            )
            if (res != 0 and ignore_zeros) or not ignore_zeros:
                result["relations"][diff_rel] = res
        return result
